/*
    Project: TCGINA
    Version: V1.30
    File   : tckeyfile.c

    Copyright (c) 2005-2012 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

/*
 * The method to derive the password from the keyfiles is taken from the
 * source code of TrueCrypt 4.0
 */

#include "tcpriv.h"
#include <tchar.h>

#pragma pack(push, 1)
typedef struct
{
    union {
        DWORD dwRemainder;
        BYTE  bRemainder[4];
    } data;
    DWORD dwZero;
} CRC32_REMAINDER;
#pragma pack(pop)

extern const DWORD rgdwCRC32Table[256];
#define InitCRC(cr)   cr.dwZero = 0; cr.data.dwRemainder = 0xffffffff
#define AddCRC(cr, b) cr.data.dwRemainder = \
  rgdwCRC32Table[cr.data.bRemainder[0] ^ (b)] ^ \
  *(DWORD *)&cr.data.bRemainder[1]
#define GetCRC(cr)    cr.data.dwRemainder

static BOOL TC_ReadKeyFile(LPSTR paszPassword, HANDLE hFile)
{
    CRC32_REMAINDER cr;
    BYTE            rgBuffer[0x10000];
    DWORD           i;
    DWORD           cRead;
    int             idx = 0;
    size_t          uTotalRead = 0;

    TC_Assert(paszPassword);
    TC_Assert(hFile != INVALID_HANDLE_VALUE);

    InitCRC(cr);

    while (ReadFile(hFile, rgBuffer, sizeof(rgBuffer), &cRead, NULL)) {
        if (cRead == 0) break;

        for (i = 0; i < cRead; i++) {
            AddCRC(cr, rgBuffer[i]);

            paszPassword[idx++] =
              (BYTE)(paszPassword[idx] + cr.data.bRemainder[3]);
            paszPassword[idx++] =
              (BYTE)(paszPassword[idx] + cr.data.bRemainder[2]);
            paszPassword[idx++] =
              (BYTE)(paszPassword[idx] + cr.data.bRemainder[1]);
            paszPassword[idx++] =
              (BYTE)(paszPassword[idx] + cr.data.bRemainder[0]);
            if (idx >= TC_MAX_PASSWORD_LEN) {
                idx = 0;
            }
            if (++uTotalRead >= 0x100000) return TRUE;
        }
    }

    return TRUE;
}

static BOOL TC_ReadKeyFileByName(LPSTR paszPassword, LPCTSTR pszKeyFile,
  BOOL fPreserveTimeStamp)
{
    HANDLE   hFile;
    HANDLE   hFile2;
    FILETIME ftCreation;
    FILETIME ftLastAccess;
    FILETIME ftLastWrite;
    BOOL     fResult;

    hFile = CreateFile(pszKeyFile, fPreserveTimeStamp ? GENERIC_READ :
      GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
      OPEN_EXISTING, 0, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
        return FALSE;
    }

    if (fPreserveTimeStamp) {
        if (!GetFileTime(hFile, &ftCreation, &ftLastAccess, &ftLastWrite)) {
            fPreserveTimeStamp = FALSE;
        }
    }

    hFile2 = CreateFile(pszKeyFile, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
      NULL, OPEN_EXISTING, 0, NULL);
    if (hFile2 != INVALID_HANDLE_VALUE) {
        fResult = TC_ReadKeyFile(paszPassword, hFile2);
        CloseHandle(hFile2);
    } else {
        fResult = FALSE;
    }

    if (fPreserveTimeStamp) {
        SetFileTime(hFile, &ftCreation, &ftLastAccess, &ftLastWrite);
    }

    CloseHandle(hFile);

    return fResult;
}

BOOL TC_ReadKeyFiles(PTC_PASSWORD pPw, LPCTSTR pszKeyFiles,
  BOOL fPreserveTimeStamps)
{
    TC_Assert(pPw);
    TC_Assert(pszKeyFiles);
    TC_Assert(*pszKeyFiles);

    for (; *pszKeyFiles; pszKeyFiles += _tcslen(pszKeyFiles) + 1) {
        DWORD dwAttr = GetFileAttributes(pszKeyFiles);
        if (dwAttr == 0xffffffff) {
            continue;
        }
        if (dwAttr & FILE_ATTRIBUTE_DIRECTORY) {
            TCHAR  szFileName[MAX_PATH];
            LPTSTR pszFilePart;
            WIN32_FIND_DATA fd;
            HANDLE hFindFile;

            if (!TC_CopyString(szFileName, MAX_PATH, pszKeyFiles)) {
                continue;
            }
            pszFilePart = szFileName + _tcslen(szFileName) - 1;
            if (*pszFilePart != TEXT('\\') && *pszFilePart != TEXT('/')) {
                if (!TC_AppendString(szFileName, MAX_PATH, TEXT("\\"))) {
                    continue;
                }
                pszFilePart++;
            }
            pszFilePart++;
            if (!TC_AppendString(szFileName, MAX_PATH, TEXT("*.*"))) {
                continue;
            }
            hFindFile = FindFirstFile(szFileName, &fd);
            if (hFindFile != INVALID_HANDLE_VALUE) {
                do {
                    if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                        continue;
                    }
                    if (!TC_CopyString(pszFilePart, MAX_PATH -
                      (DWORD)(pszFilePart - szFileName), fd.cFileName)) {
                        continue;
                    }
                    if (TC_ReadKeyFileByName(pPw->aszPassword, szFileName,
                      fPreserveTimeStamps)) {
                        pPw->nPasswordLen = TC_MAX_PASSWORD_LEN;
                    }
                } while (FindNextFile(hFindFile, &fd));
                FindClose(hFindFile);
            }
        } else {
            if (!TC_ReadKeyFileByName(pPw->aszPassword, pszKeyFiles,
              fPreserveTimeStamps)) {
                return FALSE;
            }
            pPw->nPasswordLen = TC_MAX_PASSWORD_LEN;
        }
    }

    return pPw->nPasswordLen != 0;
}