/*
    Project: TCGINA
    Version: V1.30
    File   : tcg.c

    Copyright (c) 2005-2012 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcgpriv.h"
#include "secmem.h"
#include "pwdlg.h"
#include "mediumdlg.h"
#include <tchar.h>

/*-----------------------------------------------------------------------------
// Messages                                                                  */

#define MSG_SHARED_MEMORY_ERROR            1
#define MSG_EXCLUSIVE_SESSION_RUNNING      2
#define MSG_REMOTE_SESSIONS_DISABLED       3
#define MSG_REMOTE_SESSIONS_NOT_SUPPORTED  4
#define MSG_SIMULTANEOUS_LOGON             5
#define MSG_TOO_MANY_SESSIONS              6
#define MSG_MULTIPLE_SESSIONS_WARNING      7
#define MSG_REPAIR_REDIRECTION             8
#define MSG_DISABLE_REDIRECTION_DETECTION  9
#define MSG_VOLUME_ALREADY_MOUNTED        10
#define MSG_TRUECRYPT_NOT_AVAILABLE       11
#define MSG_SERVER_CONNECTION_ERROR       12
#define MSG_REMOTE_VOLUMES_NOT_SUPPORTED  13
#define MSG_FILE_NOT_FOUND_ERROR          14
#define MSG_ERROR                         15

static int TCG_MessageBox(PTCG_CONTEXT pCtx, int iMsgNo)
{
    LPWSTR pszText;
    UINT   uStyle;

    TC_Assert(pCtx);

    switch (iMsgNo) {
    case MSG_SHARED_MEMORY_ERROR:
        pszText = L"Cannot access shared memory!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_EXCLUSIVE_SESSION_RUNNING:
        pszText =
          L"Access denied.\r\n\r\nAn interactive session with an "
          L"encrypted user profile is running.\r\n\r\n"
          L"Please try again later!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_REMOTE_SESSIONS_DISABLED:
        pszText =
          L"Access denied.\r\n\r\n"
          L"Interactive remote sessions are disabled.\r\n"
          L"(verify TCGINA registry value \"MSCount\" and \"MSOptions\")";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_REMOTE_SESSIONS_NOT_SUPPORTED:
        pszText =
          L"Access denied.\r\n\r\nInteractive remote sessions "
          L"are not supported.\r\n\r\n"
          L"Please install TCGINA FULL!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_SIMULTANEOUS_LOGON:
        pszText =
          L"Access denied.\r\n\r\nAnother user is currently "
          L"trying to log on.\r\n\r\n"
          L"Please try again later!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_TOO_MANY_SESSIONS:
        pszText =
          L"Access denied.\r\n\r\nCannot run another interactive session "
          L"with an encrypted user profile.\r\n\r\n"
          L"Please try again later!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_MULTIPLE_SESSIONS_WARNING:
        pszText =
          L"There are other interactive sessions running!\r\n"
          L"Do you really want to log on?";
        uStyle = MB_YESNOCANCEL|MB_DEFBUTTON2|MB_ICONQUESTION;
        break;

    case MSG_REPAIR_REDIRECTION:
        pszText =
          L"The location of your user profile has been changed.\r\n\r\n"
          L"Do you want to restore the location of your user profile\r\n"
          L"to its encrypted location?";
        uStyle = MB_YESNOCANCEL|MB_ICONQUESTION;
        break;

    case MSG_DISABLE_REDIRECTION_DETECTION:
        pszText =
          L"Do you want to disable profile redirection "
          L"detection for your user account?",
        uStyle = MB_YESNOCANCEL|MB_ICONQUESTION|MB_DEFBUTTON2;
        break;

    case MSG_VOLUME_ALREADY_MOUNTED:
        pszText =
          L"Volume of wrong user profile is still mounted!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_TRUECRYPT_NOT_AVAILABLE:
        pszText =
          L"TrueCrypt not available!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_SERVER_CONNECTION_ERROR:
        pszText =
          L"Cannot connect to profile server!";
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    case MSG_REMOTE_VOLUMES_NOT_SUPPORTED:
        pszText =
          L"TCGINA LITE does not support remote volumes.\r\n"
          L"Please install TCGINA FULL!",
        uStyle = MB_OK|MB_ICONSTOP;
        break;

    default:
        TC_DbgPrintfA("ERROR: Invalid message number (%d)!\n", iMsgNo);
        return IDCANCEL;
    }

    return pCtx->pWlxFuncs->WlxMessageBox(pCtx->hWlx, NULL, pszText,
      L"TCGINA", uStyle);
}

static int TCG_MessageBoxEx(PTCG_CONTEXT pCtx, int iMsgNo, LPWSTR pszText)
{
    UNREFERENCED_PARAMETER(iMsgNo);

    TC_Assert(pCtx);
    TC_Assert(pszText);

    return pCtx->pWlxFuncs->WlxMessageBox(pCtx->hWlx, NULL, pszText,
      L"TCGINA", MB_OK|MB_ICONSTOP);
}

/*-----------------------------------------------------------------------------
// Profile info                                                              */

static BOOL TCG_LogonWithStandardProfile(PTCG_CONTEXT pCtx,
  LPCWSTR pwszUserName, LPCWSTR pwszDomain)
{
    TC_Assert(pCtx);
    TC_Assert(pwszUserName);
    TC_Assert(pwszDomain);

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        BOOL fRejectSession = FALSE;
        if (pCtx->dwMSOptions & TCG_MSO_PROTECT_ENCRYPTED_SESSION) {
            PTCG_ENCRYPTED_PROFILE_INFO pepi;
            if (!TCG_LockSharedMemory()) {
                TCG_MessageBox(pCtx, MSG_SHARED_MEMORY_ERROR);
                return FALSE;
            }
            for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
                if (pepi->iStatus) {
                    if (_tcscmp(pepi->szUserName, pwszUserName) == 0 &&
                    _tcscmp(pepi->szDomain, pwszDomain) == 0) {
                        continue;
                    }
                    fRejectSession = TRUE;
                    break;
                }
            }
            TCG_UnlockSharedMemory();
        }
        if (fRejectSession) {
            TC_DbgPrintfA("Access denied (interactive session with encrpyted "
                "profile is running).\n");
            TCG_MessageBox(pCtx, MSG_EXCLUSIVE_SESSION_RUNNING);
            return FALSE;
        }
        return TRUE;
    }
#endif

    UNREFERENCED_PARAMETER(pCtx);
    UNREFERENCED_PARAMETER(pwszUserName);
    UNREFERENCED_PARAMETER(pwszDomain);

    TC_DbgPrintfA("Profile is unencrypted.\n");

    if ((pCtx->dwMSOptions & TCG_MSO_PROTECT_ENCRYPTED_SESSION) &&
      pCtx->fRemoteSession) {
#ifdef SUPPORT_SHARED_MEMORY
        TC_Assert(pCtx->dwMSCount == 0);
        TCG_MessageBox(pCtx, MSG_REMOTE_SESSIONS_DISABLED);
#else
        TCG_MessageBox(pCtx, MSG_REMOTE_SESSIONS_NOT_SUPPORTED);
#endif
        return FALSE;
    }

    return TRUE;
}

static BOOL TCG_FindProfileInfo(PTCG_CONTEXT pCtx,
  LPCWSTR pwszUserName, LPCWSTR pwszDomain, PINT piStatus,
  TCG_EPI_HANDLE *phEpi)
{
    TC_Assert(pCtx);
    TC_Assert(pwszUserName);
    TC_Assert(pwszDomain);
    TC_Assert(piStatus);
    TC_Assert(phEpi);

    *piStatus = TCG_EPIS_INVALID;
    *phEpi = NULL;

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        PTCG_ENCRYPTED_PROFILE_INFO pepi;
        if (!TCG_LockSharedMemory()) {
            TCG_MessageBox(pCtx, MSG_SHARED_MEMORY_ERROR);
            return FALSE;
        }
        for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
            if (pepi->iStatus == TCG_EPIS_AUTHENTICATED) {
                TCG_UnlockSharedMemory();
                TC_DbgPrintfA("Access denied (due to a simultaneous logon attempt).\n");
                TCG_MessageBox(pCtx, MSG_SIMULTANEOUS_LOGON);
                return FALSE;
            }
        }
        for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
            if (pepi->iStatus != TCG_EPIS_UNUSED) {
                if (wcscmp(pepi->szUserName, pwszUserName) == 0 &&
                  wcscmp(pepi->szDomain, pwszDomain) == 0) {
                    *piStatus = pepi->iStatus;
                    *phEpi = pepi;
                    break;
                }
            }
        }
        TCG_UnlockSharedMemory();
        return TRUE;
    }
#else
    UNREFERENCED_PARAMETER(pCtx);
    UNREFERENCED_PARAMETER(pwszUserName);
    UNREFERENCED_PARAMETER(pwszDomain);
#endif
    return TRUE;
}

static TCG_EPI_HANDLE TCG_ReserveProfileInfo(PTCG_CONTEXT pCtx,
  LPCWSTR pwszUserName, LPCWSTR pwszDomain, PINT piStatus)
{
    TC_Assert(pCtx);
    TC_Assert(pwszUserName);
    TC_Assert(pwszDomain);

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        PTCG_ENCRYPTED_PROFILE_INFO pepi;
        INT iPrevStatus = TCG_EPIS_INVALID;
        if (piStatus) {
            *piStatus = TCG_EPIS_INVALID;
        }
        if (!TCG_LockSharedMemory()) {
            TCG_MessageBox(pCtx, MSG_SHARED_MEMORY_ERROR);
            return NULL;
        }
        for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
            if (pepi->iStatus == TCG_EPIS_AUTHENTICATED) {
                TCG_UnlockSharedMemory();
                TCG_MessageBox(pCtx, MSG_SIMULTANEOUS_LOGON);
                return NULL;
            }
        }
        for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
            if (pepi->iStatus != TCG_EPIS_UNUSED) {
                if (wcscmp(pepi->szUserName, pwszUserName) == 0 &&
                  wcscmp(pepi->szDomain, pwszDomain) == 0) {
                    iPrevStatus = pepi->iStatus;
                    break;
                }
            }
        }
        if (pepi == pCtx->pProfileEnd) {
            for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd; pepi++) {
                if (pepi->iStatus == TCG_EPIS_UNUSED) {
                    TC_CopyString(pepi->szUserName, countof(pepi->szUserName),
                      pwszUserName);
                    TC_CopyString(pepi->szDomain, countof(pepi->szDomain),
                      pwszDomain);
                    pepi->iStatus = TCG_EPIS_AUTHENTICATED;
                    iPrevStatus = TCG_EPIS_UNUSED;
                    break;
                }
            }
        }
        TCG_UnlockSharedMemory();
        if (pepi == pCtx->pProfileEnd || pepi->iStatus == TCG_EPIS_LOGGING_OFF) {
            TC_DbgPrintfA("Access denied (too many interactive sessions).\n");
            TCG_MessageBox(pCtx, MSG_TOO_MANY_SESSIONS);
            return NULL;
        }
        if (pCtx->dwMSOptions & TCG_MSO_RUNNING_SESSIONS_WARNING) {
            PWTS_SESSION_INFO pSessions;
            DWORD nSessions;
            if (WTSEnumerateSessions(WTS_CURRENT_SERVER_HANDLE, 0, 1,
              &pSessions, &nSessions)) {
                DWORD i;
                DWORD cbSize;
                LPTSTR pBuffer;
                for (i = 0; i < nSessions; i++) {
                    if (WTSQuerySessionInformation(WTS_CURRENT_SERVER_HANDLE,
                      pSessions[i].SessionId, WTSUserName, &pBuffer, &cbSize)) {
                        if (*pBuffer) {
                            WTSFreeMemory(pBuffer);
                            break;
                        }
                        WTSFreeMemory(pBuffer);
                    }
                }
                WTSFreeMemory(pSessions);
                if (i != nSessions) {
                    TC_DbgPrintfA("Multiple interactive sessions warning.\n");
                    if (TCG_MessageBox(pCtx, MSG_MULTIPLE_SESSIONS_WARNING)
                      != IDYES) {
                        if (pepi->iStatus == TCG_EPIS_AUTHENTICATED) {
                            pepi->iStatus = TCG_EPIS_UNUSED;
                        }
                        return NULL;
                    }
                }
            }
        }
        if (piStatus) {
            *piStatus = pepi->iStatus;
        }
        return (TCG_EPI_HANDLE) pepi;
    }
#else
    UNREFERENCED_PARAMETER(pwszUserName);
    UNREFERENCED_PARAMETER(pwszDomain);
#endif

    if (pCtx->fRemoteSession) {
#ifdef SUPPORT_SHARED_MEMORY
        TCG_MessageBox(pCtx, MSG_REMOTE_SESSIONS_DISABLED);
        return NULL;
#else
        TCG_MessageBox(pCtx, MSG_REMOTE_SESSIONS_NOT_SUPPORTED);
        return NULL;
#endif
    }
    if (piStatus) {
        *piStatus = TCG_EPIS_AUTHENTICATED;
    }
    pCtx->pProfileBegin->iStatus = TCG_EPIS_AUTHENTICATED;
    return (TCG_EPI_HANDLE) pCtx->pProfileBegin;
}

static BOOL TCG_ChangeProfileInfo(PTCG_CONTEXT pCtx,
  PTCG_ENCRYPTED_PROFILE_INFO src_pepi)
{
    PTCG_ENCRYPTED_PROFILE_INFO pepi;

    TC_Assert(pCtx);

    pepi = (PTCG_ENCRYPTED_PROFILE_INFO) pCtx->hEpi;
    if (pepi == NULL) {
        return FALSE;
    }

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        if (!TCG_LockSharedMemory()) {
            if (src_pepi == NULL) {
                /* force write access */
                TC_DbgPrintfA(
                  "Shared memory is written without mutex ownership.\n");
                pepi->iStatus = TCG_EPIS_UNUSED;
                return TRUE;
            }
            TCG_MessageBox(pCtx, MSG_SHARED_MEMORY_ERROR);
            return FALSE;
        }
        if (src_pepi) {
            if (pepi->iStatus <= TCG_EPIS_LOGGING_OFF) {
                TC_DbgPrintfA("Profile is already logging/logged off.\n");
                TCG_UnlockSharedMemory();
                return FALSE;
            }
            *pepi = *src_pepi;
        } else {
            memset(pepi, '\0', sizeof(*pepi));
        }
        TCG_UnlockSharedMemory();
        return TRUE;
    }
#endif
    if (src_pepi) {
        *pepi = *src_pepi;
    } else {
        memset(pepi, '\0', sizeof(*pepi));
    }

    return TRUE;
}

static BOOL TCG_GetProfileInfo(PTCG_CONTEXT pCtx,
  PTCG_ENCRYPTED_PROFILE_INFO dest_pepi)
{
    PTCG_ENCRYPTED_PROFILE_INFO pepi;

    TC_Assert(pCtx);

    pepi = (PTCG_ENCRYPTED_PROFILE_INFO)pCtx->hEpi;

    if (pepi == NULL) {
        return FALSE;
    }

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        if (!TCG_LockSharedMemory()) {
            return FALSE;
        }
        if (dest_pepi) {
            *dest_pepi = *pepi;
        }
        TCG_UnlockSharedMemory();
        return TRUE;
    }
#endif
    if (dest_pepi) {
        *dest_pepi = *pCtx->pProfileBegin;
    }
    return TRUE;
}

static BOOL TCG_FreeProfileInfo(PTCG_CONTEXT pCtx,
  PTCG_ENCRYPTED_PROFILE_INFO dest_pepi, PDWORD pdwProtectedDrives)
{
    PTCG_ENCRYPTED_PROFILE_INFO pepi;

    TC_Assert(pCtx);

    pepi = (PTCG_ENCRYPTED_PROFILE_INFO)pCtx->hEpi;

    if (pdwProtectedDrives) {
        *pdwProtectedDrives = 0;
    }
    if (pepi == NULL) {
        return FALSE;
    }

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount) {
        if (!TCG_LockSharedMemory()) {
            /* Force access */
            if (pepi->iStatus >= TCG_EPIS_AUTHENTICATED) {
                if (pepi->iStatus == TCG_EPIS_AUTHENTICATED) {
                    if (dest_pepi) {
                        TC_DbgPrintfA(
                          "Shared memory is read without mutex ownership.\n");
                        *dest_pepi = *pepi;
                        return TRUE;
                    }
                }
                TC_DbgPrintfA(
                  "Shared memory is written without mutex ownership.\n");
                pepi->iStatus = TCG_EPIS_UNUSED;
            }
            return FALSE;
        }
        if (pdwProtectedDrives) {
            PTCG_ENCRYPTED_PROFILE_INFO pepi2;
            for (pepi2 = pCtx->pProfileBegin; pepi2 != pCtx->pProfileEnd;
              pepi2++) {
                if (pepi2 == pepi) continue;
                if (pepi2->iStatus > TCG_EPIS_AUTHENTICATED) {
                    *pdwProtectedDrives |= pepi2->dwMountedDrives;
                }
            }
        }
        if (dest_pepi) {
            *dest_pepi = *pepi;
        }
        pepi->iStatus = TCG_EPIS_LOGGING_OFF;
        TCG_UnlockSharedMemory();
        return TRUE;
    }
#endif
    if (dest_pepi) {
        *dest_pepi = *pCtx->pProfileBegin;
    }

    return TRUE;
}

/*-----------------------------------------------------------------------------
// Helper function to disable profile image path redirection repair          */

static BOOL RegSetInteger(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName, DWORD dwData)
{
    HKEY hKey;
    BOOL fResult = FALSE;

    TC_Assert(pszKeyName);
    TC_Assert(pszValueName);

    if (RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
      KEY_WRITE, NULL, &hKey, NULL) != ERROR_SUCCESS) {
        return FALSE;
    }
    if (RegSetValueEx(hKey, pszValueName, 0, REG_DWORD, (const BYTE*)&dwData,
      sizeof(DWORD)) == ERROR_SUCCESS) {
        fResult = TRUE;
    }
    RegCloseKey(hKey);

    return fResult;
}

/*-----------------------------------------------------------------------------
// Callback function for keyfile dialog box                                  */

typedef struct
{
    LPTSTR  pszFileNames;
    LPCTSTR pszDrives;
} FILE_EXISTS_DATA, *PFILE_EXISTS_DATA;

static BOOL FileOrDirectoryExists(LPCTSTR pszFileName)
{
    HANDLE hFindFile;
    WIN32_FIND_DATA fd;

    hFindFile = FindFirstFile(pszFileName, &fd);
    if (hFindFile == INVALID_HANDLE_VALUE) {
        HANDLE hFile;
        hFile = CreateFile(pszFileName, 0, FILE_SHARE_READ|FILE_SHARE_WRITE,
          NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
            CloseHandle(hFile);
            return TRUE;
        } else if (GetLastError() == ERROR_ACCESS_DENIED) {
            return TRUE;
        }

        return FALSE;
    }

    FindClose(hFindFile);

    return TRUE;
}

static BOOL CALLBACK TCG_KeyFilesExists(VOID *pCtx)
{
    LPTSTR  pszFileNames = ((PFILE_EXISTS_DATA)pCtx)->pszFileNames;
    LPCTSTR pszDrives = ((PFILE_EXISTS_DATA)pCtx)->pszDrives;

    if (*pszDrives == TEXT('\0')) {
        return TC_FileExists(pszFileNames);
    }

    while (*pszDrives) {
        TCHAR chDrive = *pszFileNames;
        LPTSTR pszTemp = pszFileNames;
        while (*pszTemp) {
            if (*pszTemp == chDrive && pszTemp[1] == TEXT(':')) {
                *pszTemp = *pszDrives;
            }
            pszTemp += _tcslen(pszTemp) + 1;
        }
        pszTemp = pszFileNames;
        while (*pszTemp) {
            if (!FileOrDirectoryExists(pszTemp)) {
                break;
            }
            pszTemp += _tcslen(pszTemp) + 1;
        }
        if (*pszTemp == TEXT('\0')) {
            return TRUE;
        }
        pszDrives++;
    }

    return FALSE;
}

/*-----------------------------------------------------------------------------
// Logon                                                                     */

BOOL WINAPI TCG_Logon(PTCG_CONTEXT pCtx, LPCWSTR pwszUserName,
  LPCWSTR pwszPassword, LPCWSTR pwszDomain)
{
    LPCTSTR                    pszSid;
    TC_HANDLE                  hTC;
    TCG_SETTINGS               ts;
    BOOL                       fProfileImagePathExists;
    PASSWORD_DIALOG_DATA       pdd;
    TCG_KEY_DATA               *ptkd;
    BOOL                       fTryLogonPassword;
    BOOL                       fResult;
    int                        iSessionType;
    TCG_ENCRYPTED_PROFILE_INFO epi;
    INT                        iStatus;

    TC_Assert(pCtx);
    TC_Assert(pCtx->pWlxFuncs);
    TC_Assert(pCtx->pWlxFuncs->WlxMessageBox);
    TC_Assert(pCtx->pWlxFuncs->WlxDialogBoxParam);
    TC_Assert(pwszUserName);

    if (pwszPassword == NULL) pwszPassword = L"";
    if (pwszDomain == NULL) pwszDomain = L"";

    iSessionType = TCG_ST_UNKNOWN;
    memset(&epi, '\0', sizeof(epi));
    ptkd = NULL;

    /* Get location of profile image path */
    pszSid = TC_GetProfileStringSid(pwszUserName);
    if (pszSid) {
        LPTSTR  pszRegKeyName;
        size_t  nSidSize;
        HKEY    hKey;
        TCHAR   szTemp[MAX_PATH];
        DWORD   dwType;
        DWORD   cbSize;
        DWORD   dwAutoRedirectionRepair;
        DWORD   dwDisableRedirectionDetection;
        BOOL    fLocationHasBeenChanged;

#ifdef SUPPORT_UNLOADING_REGISTRY
        if (_tcslen(pszSid) < countof(epi.szSID)) {
            _tcscpy(epi.szSID, pszSid);
        }
#endif

        nSidSize = _tcslen(pszSid);
        pszRegKeyName = (LPTSTR) TC_Alloc(
          (nSidSize + (PROFILE_LIST_KEY_NAME_BS_LEN+1)) * sizeof(TCHAR));
        if (pszRegKeyName == NULL) {
            TC_Free(pszSid);
            TC_DbgPutsA("Cannot get NTUSER.DAT path (out of memory)!\n");
            return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
        }

        _tcscpy(pszRegKeyName, PROFILE_LIST_KEY_NAME_BS);
        _tcscpy(pszRegKeyName + PROFILE_LIST_KEY_NAME_BS_LEN, pszSid);

        if (TC_GetRegistryString(HKEY_LOCAL_MACHINE, pszRegKeyName,
          TEXT("ProfileImagePath"), epi.szNtUserDat, MAX_PATH)) {
            TC_RemoveTerminatingBackslash(epi.szNtUserDat);
        } else {
            TC_DbgPrintf(TEXT("Cannot get NTUSER.DAT path (registry string (key=\"%s\") not found)!\n"),
              pszRegKeyName);
            TC_Free((PVOID)pszRegKeyName);
            TC_Free(pszSid);
            return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
        }

        /* Get redirection detection settings */
        dwAutoRedirectionRepair = 0;
        dwDisableRedirectionDetection = 0;
        _tcscpy(szTemp, TCG_REGKEY_NAME);
        if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, szTemp, 0, KEY_READ, &hKey) ==
          ERROR_SUCCESS) {
            TC_ReadRegistryInteger(hKey, TEXT("AutoRedirectionRepair"),
              &dwAutoRedirectionRepair);
            TC_ReadRegistryInteger(hKey, TEXT("DisableRedirectionDetection"),
              &dwDisableRedirectionDetection);
            RegCloseKey(hKey);
        }
        TC_AppendString(szTemp, MAX_PATH, TEXT("\\"));
        TC_AppendString(szTemp, MAX_PATH, pwszUserName);
        if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, szTemp, 0, KEY_READ, &hKey) ==
          ERROR_SUCCESS) {
            TC_ReadRegistryInteger(hKey, TEXT("AutoRedirectionRepair"),
              &dwAutoRedirectionRepair);
            TC_ReadRegistryInteger(hKey, TEXT("DisableRedirectionDetection"),
              &dwDisableRedirectionDetection);
            RegCloseKey(hKey);
            if (TC_ReadRegistryString(hKey, TEXT("SID"), szTemp, MAX_PATH)) {
                iSessionType = _tcscmp(szTemp, pszSid) ? TCG_ST_STANDARD :
                  TCG_ST_ENCRYPTED;
            }
        }

        /* Detect user profile redirection */
        fLocationHasBeenChanged = FALSE;
        dwType = REG_EXPAND_SZ;
        cbSize = 0;
        if (iSessionType != TCG_ST_STANDARD &&
          dwDisableRedirectionDetection == 0 &&
          RegOpenKeyEx(HKEY_LOCAL_MACHINE, szTemp, 0, KEY_READ, &hKey) ==
          ERROR_SUCCESS) {
            fLocationHasBeenChanged = FALSE;
            if (TC_ReadRegistryString(hKey, TEXT("EncryptedProfileImagePath"),
                szTemp, MAX_PATH)) {
                if (_tcsicmp(szTemp, epi.szNtUserDat) &&
                  *szTemp) {
                    cbSize = MAX_PATH*sizeof(TCHAR);
                    if (RegQueryValueEx(hKey,
                        TEXT("EncryptedProfileImagePath"), NULL, &dwType,
                        (LPBYTE)szTemp, &cbSize) == ERROR_SUCCESS) {
                        fLocationHasBeenChanged = TRUE;
                    }
                }
            }
            RegCloseKey(hKey);
        }
        TC_Free(pszSid);
        if (fLocationHasBeenChanged) {
            int iResult;
            TCG_ReadSettings(&ts, pwszUserName);
            if (dwAutoRedirectionRepair == 0 && (iResult = TCG_MessageBox(pCtx,
              MSG_REPAIR_REDIRECTION)) != IDYES) {
                fLocationHasBeenChanged = FALSE;
                if (iResult == IDNO) {
                    if (TCG_MessageBox(pCtx, MSG_DISABLE_REDIRECTION_DETECTION) == IDYES) {
                        _tcscpy(szTemp, TCG_REGKEY_NAME);
                        TC_AppendString(szTemp, MAX_PATH, TEXT("\\"));
                        TC_AppendString(szTemp, MAX_PATH, pwszUserName);
                        RegSetInteger(HKEY_LOCAL_MACHINE, szTemp,
                          TEXT("DisableRedirectionDetection"), 2);
                    }
                }
            }
            if (fLocationHasBeenChanged) {
                fLocationHasBeenChanged = FALSE;
                if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, pszRegKeyName, 0, NULL,
                  0, KEY_WRITE, NULL, &hKey, NULL) == ERROR_SUCCESS) {
                    if (RegSetValueEx(hKey, TEXT("ProfileImagePath"), 0, dwType,
                      (LPBYTE)szTemp, cbSize) == ERROR_SUCCESS) {
                        fLocationHasBeenChanged = TRUE;
                        TC_CopyString(epi.szNtUserDat, MAX_PATH,
                          szTemp);
                    }
                    RegCloseKey(hKey);
                }
            }
            if (!fLocationHasBeenChanged) {
                TC_Free((PVOID)pszRegKeyName);
                TC_DbgPutsA("Profile image path has not been restored!\n");
                return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
            }
        }
        TC_Free((PVOID)pszRegKeyName);
    } else {
        TC_DbgPutsA("Cannot get NTUSER.DAT path (string SID not available)!\n");
        return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
    }

    fProfileImagePathExists = TC_PathExists(epi.szNtUserDat);

    if (!TCG_ProfileImagePathIsValid(epi.szNtUserDat,
      pwszUserName)) {
        TCG_MessageBox(pCtx, MSG_VOLUME_ALREADY_MOUNTED);
        return FALSE;
    }

    if (!TC_AppendString(epi.szNtUserDat, MAX_PATH,
      TEXT("\\NTUSER.DAT"))) {
        *epi.szNtUserDat = TEXT('\0');
        TC_DbgPutsA("NTUSER.DAT path is too long!\n");
        return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
    }

    if (pCtx->dwOSVersion == TCG_OS_WIN_2000 && pCtx->fRemoteSession) {
        hTC = TC_Open(0);
        if (hTC) {
            TC_RefreshDriveLetters(hTC);
            TC_Close(hTC);
        }
    }

    if (TC_FileExists(epi.szNtUserDat)) {
#ifdef SUPPORT_SHARED_MEMORY
        if (!TCG_FindProfileInfo(pCtx, pwszUserName, pwszDomain, &iStatus,
          &pCtx->hEpi)) {
            return FALSE;
        }
        if (pCtx->hEpi == NULL) {
            return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
        }
        if (iStatus == TCG_EPIS_LOGGED_ON) {
            return TRUE;
        }
#else
        return TCG_LogonWithStandardProfile(pCtx, pwszUserName, pwszDomain);
#endif
    }

    TCG_DumpContext(pCtx);
    pCtx->hEpi = TCG_ReserveProfileInfo(pCtx, pwszUserName, pwszDomain,
      &iStatus);
    if (pCtx->hEpi == NULL) {
        return FALSE;
    }
    if (iStatus != TCG_EPIS_AUTHENTICATED) {
        return TRUE;
    }

    TCG_ReadSettings(&ts, pwszUserName);
    /* user profiles must not be read-only */
    ts.dwTCMountOptions &= ~TC_READ_ONLY;
    epi.dwTCOptions = ts.dwTCOptions;

#ifdef SUPPORT_NETWORK
    if (TCG_MapNetworkDrive(ts.szNDLocalName, ts.szNDRemoteName,
      ts.szNDUserName, ts.szNDPassword) == NO_ERROR) {
        TC_CopyString(epi.szNDLocalName,
          sizeof(epi.szNDLocalName)/sizeof(TCHAR), ts.szNDLocalName);
    }
#endif

    if (*ts.szTCPath && ts.szTCFileName[0] != TEXT('\\') &&
      ts.szTCFileName[0] != TEXT('/')) {
        TCHAR   szPath[MAX_PATH];
        LPCTSTR pszSearchPath = ts.szTCPath;
        for (pszSearchPath = ts.szTCPath; *pszSearchPath;
          pszSearchPath += _tcslen(pszSearchPath) + 1) {
            TC_CopyString(szPath, MAX_PATH, pszSearchPath);
            TC_AppendString(szPath, MAX_PATH, TEXT("\\"));
            TC_AppendString(szPath, MAX_PATH, ts.szTCFileName);
            if (TC_FileExists(szPath)) {
                TC_CopyString(ts.szTCFileName, countof(ts.szTCFileName),
                  szPath);
                break;
            }
        }
        if (*pszSearchPath == TEXT('\0')) {
            *ts.szTCFileName = TEXT('\0');
        }
    }

    hTC = TC_Open(0);
    if (hTC == NULL) {
        TCG_MessageBox(pCtx, MSG_TRUECRYPT_NOT_AVAILABLE);
#ifdef SUPPORT_NETWORK
        TCG_UnmapNetworkDrive(epi.szNDLocalName);
#endif
        TCG_ChangeProfileInfo(pCtx, NULL);
        pCtx->hEpi = NULL;
        TC_WipeMemory(&ts, sizeof(ts));
        TC_WipeMemory(&epi, sizeof(epi));
        goto error;
    }

    ptkd = (PTCG_KEY_DATA) sec_calloc(1, sizeof(*ptkd));
    if (ptkd == NULL) {
        goto error;
    }
#ifdef SUPPORT_KEYFILES
    if (*ts.szKFPassword) {
        if (!TC_CopyStringToUTF8(ptkd->aszPassword,
          sizeof(ptkd->aszPassword), ts.szKFPassword)) {
            *ptkd->aszPassword = '\0';
        }
    } else {
#endif
        if (!TC_CopyStringToUTF8(ptkd->aszPassword,
          sizeof(ptkd->aszPassword), pwszPassword)) {
            *ptkd->aszPassword = '\0';
        }
#ifdef SUPPORT_KEYFILES
    }
    TC_Assert(KFI_Test());
    if (!TCG_ReadKeyFile(pCtx, ptkd, &ts)) {
        goto error;
    }
    if (*ptkd->kdDecryptedData.wszHostName) {
        TC_DbgPrintf(TEXT("Volume name taken from keyfile (%s)\n"),
          ptkd->kdDecryptedData.wszHostName);
        TC_CopyString(ts.szTCFileName, countof(ts.szTCFileName),
          ptkd->kdDecryptedData.wszHostName);
    }
    if (ptkd->kdDecryptedData.chDriveLetter) {
        TC_DbgPrintf(TEXT("Drive letter taken from keyfile (%c:)\n"),
          ptkd->kdDecryptedData.chDriveLetter);
        ts.szTCDrive[0] = ptkd->kdDecryptedData.chDriveLetter;
        ts.szTCDrive[1] = TEXT(':');
        ts.szTCDrive[2] = TEXT('\\');
        ts.szTCDrive[3] = TEXT('\0');
    }
#endif
    /* Update TC volume drive letter */
    if (!fProfileImagePathExists) {
        TC_DbgPrintf(TEXT("Drive letter taken from profile path (%s)\n"),
          epi.szNtUserDat);
        ts.szTCDrive[0] = *epi.szNtUserDat;
        ts.szTCDrive[1] = TEXT(':');
        ts.szTCDrive[2] = TEXT('\\');
        ts.szTCDrive[3] = TEXT('\0');
    }

    memset(&pdd, '\0', sizeof(pdd));
    if (ts.dwTCOptions & TC_DISPLAY_PASSWORD) {
        pdd.fDisplayPassword = TRUE;
    }
    pdd.pszCaption = TEXT("Enter Volume Password");
    pdd.uCodePage = CP_ACP;
    pdd.pPassword = ptkd->aszPassword;
    pdd.cbPasswordSize = TC_MAX_PASSWORD;
    pdd.pwszTempPassword = ptkd->tmp.wszTempPassword;
    pdd.dwMinPasswordLength = 1;
    pdd.dwMaxPasswordLength = TC_MAX_PASSWORD_LEN;
    pdd.dwTimeout = ts.dwTCPasswordTimeout;
    pdd.dwTotalTimeout = ts.dwTCPasswordTotalTimeout;

    if (ptkd->kdDecryptedData.wPasswordSize == 0) {
        if (!TC_CopyStringToAnsi(ptkd->aszPassword,
          sizeof(ptkd->aszPassword), pwszPassword)) {
            *ptkd->aszPassword = '\0';
        }
    }

    fTryLogonPassword = *ptkd->aszPassword != '\0';
    if (ptkd->kdDecryptedData.wPasswordSize) {
        fTryLogonPassword =
          (ptkd->kdDecryptedData.wPasswordType & 3) != KFIPT_INCOMPLETE;
    } else if (ts.dwTCOptions & TC_IGNORE_USER_ACCOUNT_PASSWORD) {
        fTryLogonPassword = FALSE;
    }

    if (TC_RemoteFile(ts.szTCFileName)) {
#ifdef SUPPORT_NETWORK
        TCHAR       szUserName[MAX_PATH];
        TCHAR       szDomainAndUserName[MAX_PATH];
        TCHAR       szRemoteName[MAX_PATH];
        NETRESOURCE nr;
        DWORD       dwErrorCode;

        TC_CopyStringFromUnicode(szUserName, countof(szUserName), pwszUserName);
        dwErrorCode = ERROR_INVALID_PARAMETER;
        if (_tcschr(szUserName, TEXT('\\')) == 0) {
            if (TC_CopyString(szDomainAndUserName, MAX_PATH,
              ts.szTCFileName+2)) {
                LPTSTR pszTemp = _tcschr(szDomainAndUserName, TEXT('\\'));
                if (pszTemp) {
                    pszTemp[1] = TEXT('\0');
                    if (TC_AppendString(szDomainAndUserName, MAX_PATH,
                      szUserName)) {
                        dwErrorCode = NO_ERROR;
                    }
                }
            }
        }

        if (dwErrorCode == NO_ERROR) {
            LPTSTR pszTemp;

            TC_CopyString(szRemoteName, MAX_PATH, ts.szTCFileName);
            pszTemp = _tcschr(szRemoteName + 2, TEXT('\\'));
            if (pszTemp) *pszTemp = TEXT('\0');

            TC_ZeroMemory(&nr, sizeof(nr));
            nr.dwType = RESOURCETYPE_DISK;
            nr.lpRemoteName = szRemoteName;

            dwErrorCode = WNetAddConnection2(&nr, pwszPassword,
              szDomainAndUserName, 0);
        }
        if (dwErrorCode != NO_ERROR) {
            TC_DbgPrintfA("WNetAddConnection2 failed (ErrorCode=%d)\n",
              dwErrorCode);
            TCG_MessageBox(pCtx, MSG_SERVER_CONNECTION_ERROR);
            goto error;
        }

        TC_CopyString(epi.szRemoteName, countof(epi.szRemoteName),
          szRemoteName);
#else
        TCG_MessageBox(pCtx, MSG_REMOTE_VOLUMES_NOT_SUPPORTED);
        goto error;
#endif
    }

    fResult = TRUE;

    if (*ts.szTCKeyFileNames) {
        /* Search for keyfiles */
        FILE_EXISTS_DATA fed;
        TCHAR            szDrives[26+1+1];
        fed.pszDrives    = szDrives;
        fed.pszFileNames = ts.szTCKeyFileNames;
        if (ts.szTCKeyFileNames[1] == TEXT(':')) {
            szDrives[0] = ts.szTCKeyFileNames[0];
            if (!TC_CopyString(szDrives + 1, 26+1, ts.szTCKeyFileDrives)) {
                *szDrives = TEXT('\0');
            }
        } else {
            *szDrives = TEXT('\0');
        }
        if (!TCG_KeyFilesExists(&fed)) {
            MEDIUM_DIALOG_DATA mdd;
            memset(&mdd, '\0', sizeof(mdd));
            mdd.pszCaption = TEXT("Keyfiles");
            mdd.pszText = TEXT("Insert keyfile medium!");
            mdd.dwTimeout = ts.dwTCKeyFileTimeout;
            mdd.pfMediumAvailable = TCG_KeyFilesExists;
            mdd.pContext = &fed;
            if (pCtx->pWlxFuncs->WlxDialogBoxParam(pCtx->hWlx,
              g_hinstDLL, MAKEINTRESOURCE(IDD_MEDIUM), NULL,
              MediumDialogProc, (LPARAM)&mdd) != IDOK) {
                goto error;
            }
        }
    }

    if (TC_VolumeIsAccessable(*ts.szTCDrive)) {
        /* Dismount volume for user profile volume */
        DWORD dwProtectedDrives = pCtx->dwProtectedDrives;
#ifdef SUPPORT_SHARED_MEMORY
        if (pCtx->dwMSCount) {
            if (TCG_LockSharedMemory()) {
                PTCG_ENCRYPTED_PROFILE_INFO pepi;
                for (pepi = pCtx->pProfileBegin; pepi != pCtx->pProfileEnd;
                  pepi++) {
                    if (pepi->iStatus > TCG_EPIS_AUTHENTICATED) {
                        dwProtectedDrives |= pepi->dwMountedDrives;
                    }
                }
                TCG_UnlockSharedMemory();
            }
        }
#endif
        if ((dwProtectedDrives & TC_GetDriveBitMask(*ts.szTCDrive)) == 0) {
            TC_Dismount(hTC, ts.szTCDrive, TC_DISMOUNT_FORCIBLY);
        }

#if 0
        /* Alternative block used with persistent volumes */
        DWORD dwMountedDrives;
        if (!TC_GetMountedDrives(hTC, &dwMountedDrives, TCMD_DISMOUNTABLE)) {
            goto error;
        }
        if ((TC_GetDriveBitMask(*ts.szTCDrive) & dwMountedDrives) == 0) {
            TC_DbgPrintf(TEXT("Error: Drive letter %s is already occupied ")
              TEXT("(but not by a dismountable TrueCrypt volume)!"),
              ts.szTCDrive);
            goto error;
        }
        TC_Dismount(hTC, ts.szTCDrive, TC_DISMOUNT_FORCIBLY);
#endif
    }

    if (TC_VolumeHostExists(ts.szTCFileName)) {

        /* Try to mount TC volume with Windows user password */
        if (fTryLogonPassword) {
            if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive,
              pwszUserName, epi.szNtUserDat, ptkd, ts.dwTCMountOptions,
              ts.szTCKeyFileNames, FALSE, NULL)) {
                goto exit;
            }
        }

        /* Password dialog */
        while (pCtx->pWlxFuncs->WlxDialogBoxParam(pCtx->hWlx, g_hinstDLL,
          MAKEINTRESOURCE(IDD_PASSWORD), NULL, PasswordDialogProc,
          (LPARAM)&pdd) == IDOK) {
            if (*ptkd->aszPassword) {
                if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive,
                  pwszUserName, epi.szNtUserDat, ptkd,
                  ts.dwTCMountOptions, ts.szTCKeyFileNames, FALSE, NULL)) {
                    goto exit;
                } else {
                    DWORD dwErrorCode = TC_GetLastError(hTC);
                    if (dwErrorCode != ERROR_TC_WRONG_PASSWORD &&
                      dwErrorCode != ERROR_TC_INVALID_PASSWORD &&
                      dwErrorCode != ERROR_TC_INVALID_KEYFILES) {
                        goto display_error;
                    }
                }
            }
        }

    } else if (!ts.fTCFileNameAvailable) {

        TC_PARTITION_DATA pd;

        /* Try to mount partition with Windows user password */
        if (fTryLogonPassword) {
            if (TC_GetFirstPartition(&pd)) {
                do {
                    if (TC_GetPartitionDeviceName(&pd, ts.szTCFileName,
                        countof(ts.szTCFileName))) {
                        if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive,
                          pwszUserName, epi.szNtUserDat, ptkd,
                          ts.dwTCMountOptions, ts.szTCKeyFileNames, FALSE,
                          NULL)) {
                            goto exit;
                        }
                    }
                } while (TC_GetNextPartition(&pd));
            }
        }

        /* Password dialog */
        while (pCtx->pWlxFuncs->WlxDialogBoxParam(pCtx->hWlx, g_hinstDLL,
          MAKEINTRESOURCE(IDD_PASSWORD), NULL, PasswordDialogProc,
          (LPARAM)&pdd) == IDOK) {
            if (*ptkd->aszPassword) {
                if (TC_GetFirstPartition(&pd)) {
                    do {
                        if (TC_GetPartitionDeviceName(&pd, ts.szTCFileName,
                          countof(ts.szTCFileName))) {
                            if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive,
                              pwszUserName, epi.szNtUserDat, ptkd,
                              ts.dwTCMountOptions, ts.szTCKeyFileNames, FALSE,
                              NULL)) {
                                goto exit;
                            }
                        }
                    } while (TC_GetNextPartition(&pd));
                }
            }
        }
    } else {
        TCHAR szText[1024];
        _sntprintf(szText, countof(szText), TEXT("Error: Cannot find \"%s\""),
          ts.szTCFileName);
        TCG_MessageBoxEx(pCtx, MSG_FILE_NOT_FOUND_ERROR, szText);
    }

    goto error;

display_error:
    {
        TCHAR szText[1024];
        _tcscpy(szText, TEXT("Error: "));
        TC_GetErrorString(TC_GetLastError(hTC), szText + 7,
          countof(szText) - 7);
        TCG_MessageBoxEx(pCtx, MSG_ERROR, szText);
    }

error:
    fResult = FALSE;

exit:
    if (fResult) {
        HANDLE hFile;

        epi.dwMountedDrives |= TC_GetDriveBitMask(*ts.szTCDrive);

        TCG_AutoMountPartitions(hTC, *ts.szTCFirstAutoMountDrive, ptkd,
          ts.dwTCMountOptions, ts.szTCKeyFileNames, FALSE, NULL,
          &epi.dwMountedDrives);

        hFile = CreateFile(epi.szNtUserDat, 0,
            FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
            FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
            CloseHandle(hFile);
            epi.fWaitForNtUserDat = TRUE;
        } else {
            epi.fWaitForNtUserDat = FALSE;
        }

#ifdef SUPPORT_SERVICES
        TCG_StopServices(ts.szTCStopServices);
        TCG_StartServices(ts.szTCStartServices);
#endif
        TC_CopyString(epi.szUserName, countof(epi.szUserName), pwszUserName);
        TC_CopyString(epi.szDomain, countof(epi.szDomain), pwszDomain);

        epi.iStatus = TCG_EPIS_LOGGED_ON;
        if (!TCG_ChangeProfileInfo(pCtx, &epi)) {
            fResult = FALSE;
        }
    }
    sec_free(ptkd);
    sec_heap_compact();
    if (hTC) {
        TC_Close(hTC);
    }
    TC_WipeMemory(&ts, sizeof(ts));
    if (!fResult) {
#ifdef SUPPORT_NETWORK
        if (*epi.szRemoteName) {
            DWORD dwErrorCode;
            dwErrorCode = WNetCancelConnection2(epi.szRemoteName, 0, FALSE);
            if (dwErrorCode != NO_ERROR) {
                TC_DbgPrintf(
                  TEXT("WNetCancelConnection2 failed (ErrorCode=%d)\n"),
                    dwErrorCode);
            }
        }
        TCG_UnmapNetworkDrive(epi.szNDLocalName);
#endif
        TCG_ChangeProfileInfo(pCtx, NULL);
        pCtx->hEpi = NULL;
        TC_WipeMemory(&ts, sizeof(ts));
        TC_WipeMemory(&epi, sizeof(epi));
    }

    TCG_DumpContext(pCtx);
    TCG_DumpHeap();

    return fResult;
}

/*-----------------------------------------------------------------------------
// Token privileges                                                          */

#ifdef SUPPORT_UNLOADING_REGISTRY

static BOOL SetPrivilege(LPCTSTR szPrivilege)
{
    HANDLE           hToken;
    LUID             luid;
    TOKEN_PRIVILEGES tpNewState;

    TC_Assert(szPrivilege);

    hToken = NULL;
    if (!OpenProcessToken(GetCurrentProcess(),
      TOKEN_ADJUST_PRIVILEGES|TOKEN_QUERY, &hToken)) {
        return FALSE;
    }

    if (!LookupPrivilegeValue(NULL, szPrivilege, &luid)) {
        CloseHandle(hToken);
        return FALSE;
    }

    tpNewState.PrivilegeCount = 1;
    tpNewState.Privileges[0].Luid = luid;
    tpNewState.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    if (!AdjustTokenPrivileges(hToken, FALSE, &tpNewState, 0, NULL, NULL)) {
        CloseHandle(hToken);
        return FALSE;
    }

    CloseHandle(hToken);

    return TRUE;
}

#endif

/*-----------------------------------------------------------------------------
// Logoff                                                                    */

VOID WINAPI TCG_Logoff(PTCG_CONTEXT pCtx)
{
    TC_HANDLE                  hTC;
    TCG_ENCRYPTED_PROFILE_INFO epi;
    DWORD                      dwProtectedDrives;

    TC_Assert(pCtx);

    TC_DbgPutsA("Logoff\n");

    if (pCtx->hEpi == NULL) {
        return;
    }

    TCG_DumpContext(pCtx);

#ifdef SUPPORT_SHARED_MEMORY
    if (pCtx->dwMSCount &&
      TCG_GetProfileInfo(pCtx, &epi) && epi.iStatus > TCG_EPIS_AUTHENTICATED) {
        /* Find another session which shares our profile */
        DWORD  dwSessionId;
        if (ProcessIdToSessionId(GetCurrentProcessId(), &dwSessionId)) {
            PWTS_SESSION_INFO pSessions;
            DWORD  nSessions;
            if (WTSEnumerateSessions(WTS_CURRENT_SERVER_HANDLE, 0, 1,
              &pSessions, &nSessions)) {
                DWORD  i;
                LPTSTR pBuffer;
                DWORD  cbSize;
                for (i = 0; i < nSessions; i++) {
                    if (pSessions[i].SessionId == dwSessionId) {
                        continue;
                    }
                    if (WTSQuerySessionInformation(WTS_CURRENT_SERVER_HANDLE,
                      pSessions[i].SessionId, WTSUserName, &pBuffer,
                      &cbSize)) {
                        if (_tcscmp(pBuffer, epi.szUserName) != 0) {
                            WTSFreeMemory(pBuffer);
                            continue;
                        }
                        WTSFreeMemory(pBuffer);
                    }
                    if (WTSQuerySessionInformation(WTS_CURRENT_SERVER_HANDLE,
                      pSessions[i].SessionId, WTSDomainName, &pBuffer,
                      &cbSize)) {
                        if (_tcscmp(pBuffer, epi.szDomain) != 0) {
                            WTSFreeMemory(pBuffer);
                            continue;
                        }
                        WTSFreeMemory(pBuffer);
                    }
                    break;
                }
                WTSFreeMemory(pSessions);
                if (i != nSessions) {
                    TC_DbgPrintfA("Another session with same user profile found!\n");
                    pCtx->hEpi = NULL;
                    TC_WipeMemory(&epi, sizeof(epi));
                    return;
                }
            }
        }
    }
#endif

    if (TCG_FreeProfileInfo(pCtx, &epi, &dwProtectedDrives)) {
        hTC = TC_Open(0);
        if (hTC) {
            DWORD dwMountedDrives;
            LONG lResult;

            if ((epi.dwTCOptions & TC_DISABLE_DISMOUNTING_ON_LOGOFF) == 0) {
                TC_DbgPutsA("Flush HKEY_USERS\n");
                lResult = RegFlushKey(HKEY_USERS);
                if (lResult != ERROR_SUCCESS) {
                    TC_DbgPrintfA("Unsuccessful: Win32 Error Code 0x%04x\n", lResult);
                }
#ifdef SUPPORT_UNLOADING_REGISTRY
                if (1) {
                    HKEY hKey;
                    SetPrivilege(SE_BACKUP_NAME);
                    SetPrivilege(SE_RESTORE_NAME);
                    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                      TEXT("SYSTEM\\CurrentControlSet\\Control\\hivelist"), 0,
                      KEY_READ, &hKey) == ERROR_SUCCESS) {
                        DWORD dwIndex;
                        DWORD dwSize;
                        TCHAR szValueName[260];
                        dwIndex = 0;
                        dwSize  = countof(szValueName);
                        TC_DbgPrintf(TEXT("Searching for registry keys (with SID = %s)\n"), epi.szSID);
                        while (RegEnumValue(hKey, dwIndex++, szValueName, &dwSize,
                          NULL, NULL, NULL, NULL) == ERROR_SUCCESS) {
                            if (_tcsnicmp(szValueName, TEXT("\\REGISTRY\\USER\\"), 15) ==
                              0) {
                                if (_tcsnicmp(szValueName + 15, epi.szSID, _tcslen(epi.szSID)) ==
                                  0) {
                                    TC_DbgPrintf(TEXT("Unloading registry key: HKEY_USERS\\%s\n"), szValueName+15);
                                    lResult = RegUnLoadKey(HKEY_USERS, szValueName+15);
                                    if (lResult != ERROR_SUCCESS) {
                                        TC_DbgPrintfA("Unsuccessful: Win32 Error Code 0x%04x\n", lResult);
                                    }
                                }
                            }
                            dwSize = countof(szValueName);
                        }
                        RegCloseKey(hKey);
                    }
                }
#endif
                if (epi.fWaitForNtUserDat &&
                  TC_FileExists(epi.szNtUserDat)) {
                    /* Wait for exclusive NTUSER.DAT access */
                    DWORD  nAttempts;
                    nAttempts = 10;
                    do {
                        HANDLE hFile;
                        hFile = CreateFile(epi.szNtUserDat, GENERIC_WRITE,
                          0, 0, OPEN_EXISTING, 0, NULL);
                        if (hFile != INVALID_HANDLE_VALUE) {
                            CloseHandle(hFile);
                            break;
                        }
                        TC_DbgPutsA("Waiting for NTUSER.DAT\n");
                        Sleep(500);
                    } while (--nAttempts);
                }

                dwProtectedDrives |= pCtx->dwProtectedDrives;
                if (!TC_GetMountedDrives(hTC, &dwMountedDrives, TCMD_DISMOUNTABLE)) {
                    dwMountedDrives = 0x3ffffff;
                }
                dwMountedDrives |= epi.dwMountedDrives;
                dwMountedDrives &= ~dwProtectedDrives;
                if (dwMountedDrives) {
                    INT   i;
                    TCHAR szDrive[3] = TEXT("A:");
                    for (i = 0; i < 26; i++) {
                        if (dwMountedDrives & (1<<i)) {
                            szDrive[0] = TEXT('A') + (TCHAR)i;
                            TC_Dismount(hTC, szDrive, TC_DISMOUNT_FORCIBLY);
                        }
                    }
                }
            } else {
                pCtx->fFlushRegistry = TRUE;
            }
            if ((epi.dwTCOptions & TC_DISABLE_WIPING_ON_LOGOFF) == 0) {
                TC_WipeCache(hTC);
            }

            TC_Close(hTC);
        }
#ifdef SUPPORT_NETWORK
        if (*epi.szRemoteName) {
            DWORD dwErrorCode;
            dwErrorCode = WNetCancelConnection2(epi.szRemoteName, 0, FALSE);
            if (dwErrorCode != NO_ERROR) {
                TC_DbgPrintf(
                  TEXT("WNetCancelConnection2 failed (ErrorCode=%d)\n"),
                    dwErrorCode);
            }
        }
        TCG_UnmapNetworkDrive(epi.szNDLocalName);
#endif
        TCG_ChangeProfileInfo(pCtx, NULL);
        TC_WipeMemory(&epi, sizeof(epi));
    }
    pCtx->hEpi = NULL;
}

VOID WINAPI TCG_LoggedOff(PTCG_CONTEXT pCtx)
{
    if (pCtx->hEpi) {
        TCG_Logoff(pCtx);
    }
}
