/*
    Project: TCGINA
    Version: V1.30
    File   : tcgservices.c

    Copyright (c) 2005-2012 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcgpriv.h"
#include <tchar.h>

#define SERVICE_STOP_TIMEOUT  180000
#define SERVICE_START_TIMEOUT 180000

static BOOL WINAPI _StopService(SC_HANDLE hSCM,
  LPCTSTR pszServiceName)
{
    SC_HANDLE      hService;
    SERVICE_STATUS ssStatus;

    DWORD          dwPrevCheckPoint;
    DWORD          dwStartTickCount;
    DWORD          dwMaxTime;

    TC_DbgPrintf(TEXT("StopService(\"%s\")\n"), pszServiceName);

    hService = OpenService(hSCM, pszServiceName, SERVICE_ALL_ACCESS);

    if (!hService) {
        TC_DbgPrintfA("Cannot open service!\n");
        return FALSE;
    }

    if (!QueryServiceStatus(hService, &ssStatus)) {
        TC_DbgPrintfA("Cannot query service status!\n");
        CloseServiceHandle(hService);
        return FALSE;
    }
    if (ssStatus.dwCurrentState != SERVICE_STOP_PENDING &&
      ssStatus.dwCurrentState != SERVICE_STOPPED) {

        LPENUM_SERVICE_STATUS lpServices;
        DWORD                 cbBytesNeeded;
        DWORD                 cbServicesReturned;
        DWORD                 i;

        if (!EnumDependentServices(hService, SERVICE_ACTIVE, NULL, 0,
          &cbBytesNeeded, &cbServicesReturned) && GetLastError() ==
          ERROR_MORE_DATA) {
            lpServices = TC_Alloc(cbBytesNeeded);
            if (lpServices == NULL) {
                TC_DbgPrintfA("Out of memory!\n");
                CloseServiceHandle(hService);
                return FALSE;
            }
            if (!EnumDependentServices(hService, SERVICE_ACTIVE,
              lpServices, cbBytesNeeded, &cbBytesNeeded,
              &cbServicesReturned)) {
                TC_Free(lpServices);
                TC_DbgPrintfA("Cannot enumerate dependent services!\n");
                CloseServiceHandle(hService);
                return FALSE;
            }
            for (i = 0; i < cbServicesReturned; i++) {
                if (!_StopService(hSCM, lpServices[i].lpServiceName)) {
                    TC_Free(lpServices);
                    TC_DbgPrintfA("Cannot stop dependent service!\n");
                    CloseServiceHandle(hService);
                    return FALSE;
                }
            }
            TC_Free(lpServices);
        }

        if (!ControlService(hService, SERVICE_CONTROL_STOP, &ssStatus)) {
            TC_DbgPrintfA("Cannot stop service!\n");
            CloseServiceHandle(hService);
            return FALSE;
        }
    }

    dwPrevCheckPoint = ssStatus.dwCheckPoint;
    dwStartTickCount = GetTickCount();
    dwMaxTime = dwStartTickCount + SERVICE_STOP_TIMEOUT;
    while (ssStatus.dwCurrentState == SERVICE_STOP_PENDING) {
        DWORD dwWait = ssStatus.dwWaitHint / 10;
        if (dwWait < 1000) {
            dwWait = 1000;
        } else if (dwWait > 10000) {
            dwWait = 10000;
        }
        Sleep(dwWait);
        if (!QueryServiceStatus(hService, &ssStatus)) {
            TC_DbgPrintfA("Cannot query service status!\n");
            CloseServiceHandle(hService);
            return FALSE;
        }
        if (ssStatus.dwCurrentState != SERVICE_STOP_PENDING) {
            break;
        }
        if (ssStatus.dwCheckPoint > dwPrevCheckPoint) {
            dwStartTickCount = GetTickCount();
            dwPrevCheckPoint = ssStatus.dwCheckPoint;
        } else if (GetTickCount() - dwStartTickCount >
          ssStatus.dwWaitHint && ssStatus.dwWaitHint) {
            TC_DbgPrintfA("Service stop hangs!\n");
            CloseServiceHandle(hService);
            return FALSE;
        }
        if (GetTickCount() > dwMaxTime) {
            TC_DbgPrintfA("Time out!\n");
            CloseServiceHandle(hService);
            return FALSE;
        }
    }

    if (ssStatus.dwCurrentState != SERVICE_STOPPED) {
        TC_DbgPrintfA("Service is not stopped!\n");
        CloseServiceHandle(hService);
        return FALSE;
    }

    CloseServiceHandle(hService);

    return TRUE;
}

BOOL WINAPI TCG_StopService(LPCTSTR pszServiceName)
{
    SC_HANDLE hSCM;
    BOOL      fResult;

    hSCM = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!hSCM) {
        TC_DbgPrintfA("Cannot open SCM!\n");
        return FALSE;
    }

    fResult = _StopService(hSCM, pszServiceName);

    CloseServiceHandle(hSCM);

    return fResult;
}

BOOL WINAPI TCG_StartService(LPCTSTR pszServiceName)
{
    SC_HANDLE      hSCM;
    SC_HANDLE      hService;
    SERVICE_STATUS ssStatus;
    DWORD          dwPrevCheckPoint;
    DWORD          dwStartTickCount;
    DWORD          dwMaxTime;

    TC_DbgPrintf(TEXT("StartService(\"%s\")\n"), pszServiceName);

    hSCM = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!hSCM) {
        TC_DbgPrintfA("Cannot open SCM!\n");
        return FALSE;
    }

    hService = OpenService(hSCM, pszServiceName, SERVICE_ALL_ACCESS);
    if (!hService) {
        TC_DbgPrintfA("Cannot open service!\n");
        CloseServiceHandle(hSCM);
        return FALSE;
    }

    if (!QueryServiceStatus(hService, &ssStatus)) {
        TC_DbgPrintfA("Cannot query service status!\n");
        goto error;
    }
    if (ssStatus.dwCurrentState != SERVICE_START_PENDING &&
      ssStatus.dwCurrentState != SERVICE_RUNNING) {
        if (!StartService(hService, 0, NULL)) {
            TC_DbgPrintfA("Cannot start service!\n");
            goto error;
        }
        if (!QueryServiceStatus(hService, &ssStatus)) {
            TC_DbgPrintfA("Cannot query service status!\n");
            goto error;
        }
    }

    dwPrevCheckPoint = ssStatus.dwCheckPoint;
    dwStartTickCount = GetTickCount();
    dwMaxTime        = dwStartTickCount + SERVICE_START_TIMEOUT;
    while (ssStatus.dwCurrentState == SERVICE_START_PENDING) {
        DWORD dwWait = ssStatus.dwWaitHint / 10;
        if (dwWait < 1000) {
            dwWait = 1000;
        } else if (dwWait > 10000) {
            dwWait = 10000;
        }
        Sleep(dwWait);
        if (!QueryServiceStatus(hService, &ssStatus)) {
            TC_DbgPrintfA("Cannot query service status!\n");
            goto error;
        }
        if (ssStatus.dwCurrentState != SERVICE_START_PENDING) {
            break;
        }
        if (ssStatus.dwCheckPoint > dwPrevCheckPoint) {
            dwStartTickCount = GetTickCount();
            dwPrevCheckPoint = ssStatus.dwCheckPoint;
        } else if (GetTickCount() - dwStartTickCount >
          ssStatus.dwWaitHint && ssStatus.dwWaitHint) {
            TC_DbgPrintfA("Service start hangs!\n");
            goto error;
        }
        if (GetTickCount() > dwMaxTime) {
            TC_DbgPrintfA("Time out!\n");
            goto error;
        }
    }

    if (ssStatus.dwCurrentState != SERVICE_RUNNING) {
        TC_DbgPrintfA("Service is not running!");
        goto error;
    }

    CloseServiceHandle(hService);
    CloseServiceHandle(hSCM);

    return TRUE;

error:
    CloseServiceHandle(hService);
    CloseServiceHandle(hSCM);

    return FALSE;
}

BOOL WINAPI TCG_StopServices(LPCTSTR pszServiceNames)
{
    BOOL fResult = TRUE;

    while (*pszServiceNames) {
        if (!TCG_StopService(pszServiceNames)) {
            fResult = FALSE;
        }
        pszServiceNames += _tcslen(pszServiceNames) + 1;
    }

    return fResult;
}

BOOL WINAPI TCG_StartServices(LPCTSTR pszServiceNames)
{
    BOOL fResult = TRUE;

    while (*pszServiceNames) {
        if (!TCG_StartService(pszServiceNames)) {
            fResult = FALSE;
        }
        pszServiceNames += _tcslen(pszServiceNames) + 1;
    }

    return fResult;
}
